from abc import ABC, abstractmethod


class Transform(ABC):
    def __init__(self, keys):
        assert keys
        if isinstance(keys, str):
            keys = (keys,)
        self.keys = keys

    def __call__(self, data):
        return self._call_fun(data)

    @abstractmethod
    def _call_fun(self, data):
        ...



import os

import dicom2nifti as d2n


def nifti2dicom(input_file, output_dir, accession_number=1):
    """
    Require nifti2dicom.
    `sudo apt install nifti2dicom`
    https://github.com/biolab-unige/nifti2dicom
    Args:
        input_file (str or pathlib.Path):
        output_dir (str or pathlib.Path):
        accession_number (int, optional, default=1):

    Returns:

    """

    cmd = f'nifti2dicom -i {input_file} -o {output_dir} -a {accession_number}'
    result = os.popen(cmd)
    return result.readlines()


def dicom2nifti(input_dicom_series, output_nifti_file, pydicom_read_force=False):
    if pydicom_read_force:
        d2n.settings.pydicom_read_force = pydicom_read_force
    d2n.convert_directory(input_dicom_series, output_nifti_file)



from abc import ABC

import numpy as np

from jbag.transforms import Transform


class ImageAugmentTransform(Transform, ABC):
    def __init__(self, keys, prob):
        super().__init__(keys)
        self.prob = prob

    def __call__(self, data):
        if np.random.random() < self.prob:
            data = self._call_fun(data)
        return data


class RandFlip(ImageAugmentTransform):
    def __init__(self, keys, prob, spatial_axis=None):
        """
        Flip image along specific spatial axis.

        Args:
            keys (str or sequence):
            prob (float):
            spatial_axis (int or sequence or None): Flip the array along the assigned axis. If more than one axes are
                given, flip the array along each axis. If None, flip the data along all axes.
        """

        super().__init__(keys, prob)
        self.axis = spatial_axis

    def _call_fun(self, data):
        for key in self.keys:
            image = data[key]
            flipped_elem_data = np.flip(image, axis=self.axis)
            data[key] = flipped_elem_data.copy()
        return data


class RandRotate90(ImageAugmentTransform):
    def __init__(self, keys, prob, max_k=1, spatial_axes: tuple[int, int] = (0, 1)):
        """
        Rotate array by 90 degrees in the plan specified by spatial axes. For a 2-D array with `k = 1` and
        `axes = (0, 1)`, the rotation direction will be counterclockwise.

        Args:
            keys (str or sequence):
            prob (float):
            max_k (int, optional, default=1):
            spatial_axes (sequence, optional, default=(0, 1)): Two int numbers, rotate the array in the plane
                determined by these axes.
        """

        super().__init__(keys, prob)
        self.max_k = max_k
        self.axes = spatial_axes

    def _call_fun(self, data):
        k = np.random.randint(1, self.max_k + 1)
        for key in self.keys:
            image = data[key]
            rotated_elem_data = np.rot90(image, k=k, axes=self.axes)
            data[key] = rotated_elem_data.copy()
        return data


class RandScaleIntensity(ImageAugmentTransform):
    def __init__(self, keys, prob, factors: tuple[float, float] | float):
        """
        Scale the image intensity using factor by `i = i * (1 + factor)`.

        Args:
            keys (str or sequence):
            prob (float):
            factors (float or sequence): If provided factors is a number, the factor will be uniformly sampled from
                [-factors, factors). If factors is a pair of numbers, the factor will be picked from
                [min(factors), max(factors)].
        """

        super().__init__(keys, prob)
        if isinstance(factors, (int, float)):
            self.factors = (min(-factors, factors), max(-factors, factors))
        elif len(factors) != 2:
            raise ValueError(f'factors should be a number or a pair of numbers, got {factors}')
        else:
            self.factors = (min(factors), max(factors))

    def _call_fun(self, data):
        factor = np.random.uniform(self.factors[0], self.factors[1])
        for key in self.keys:
            image = data[key]
            scaled_data = image * (1. + factor)
            data[key] = scaled_data
        return data


class RandShiftIntensity(ImageAugmentTransform):
    def __init__(self, keys, prob, offsets: tuple[float, float] | float):
        """
        Scale the image intensity using factor by `i = i * (1 + factor)`.

        Args:
            keys (str or sequence):
            prob (float):
            offsets (float or sequence): If provided offset is a number, the offset will be uniformly picked from
                [-offsets, offsets). If offsets is a pair of numbers, the offsets will be picked from
                [min(offsets), max(offsets)].
        """

        super().__init__(keys, prob)
        if isinstance(offsets, (int, float)):
            self.offsets = (min(-offsets, offsets), max(-offsets, offsets))
        elif len(offsets) != 2:
            raise ValueError(f'offsets should be a number or a pair of numbers, got {offsets}')
        else:
            self.factors = (min(offsets), max(offsets))

    def _call_fun(self, data):
        offset = np.random.uniform(self.offsets[0], self.offsets[1])
        for key in self.keys:
            image = data[key]
            scaled_data = image + offset
            data[key] = scaled_data
        return data



import re
import tomllib

import lazyConfig


def get_config(file):
    with open(file, 'rb') as f:
        config = tomllib.load(f)
    refine_nodes(config, config)
    config = lazyConfig.from_primitive(config)
    return config


def refine_nodes(node, root):
    for k, v in node.items():
        if isinstance(v, str):
            replace_str(k, node, root)
        elif isinstance(v, dict):
            refine_nodes(v, root)
        elif isinstance(v, list):
            for i, inner_v in enumerate(v):
                if isinstance(inner_v, str):
                    replace_str(k, node, root, list_index=i)
                elif isinstance(inner_v, dict):
                    refine_nodes(inner_v, root)


def replace_str(key, node, root, list_index=None):
    """
    If `node[key]` is list type, set list_index to the index of current element of `node[key]`.

    Args:
        key (str):
        node (mapping):
        root (mapping):
        list_index (int or None, optional, default=None):

    Returns:

    """

    reference_patten = r'(\$\{.+?\})'
    reference_key_patten = r'\$\{(.+)\}'
    if list_index is not None:
        value = node[key][list_index]
    else:
        value = node[key]
    match = re.findall(reference_patten, value)
    if not match:
        return
    for each in match:
        replace_key = re.search(reference_key_patten, each).group(1)
        replace_node, replace_key = get_node_key(replace_key, node, root)
        if isinstance(replace_node[replace_key], str):
            replace_str(replace_key, replace_node, root)
        if list_index is not None:
            node[key][list_index] = node[key][list_index].replace(each, root[replace_key])
        else:
            node[key] = node[key].replace(each, str(replace_node[replace_key]))


def get_node_key(key, node, root):
    """
    Get the inner node and key.

    Args:
        key (str):
        node (dict):
        root (mapping):

    Returns:

    """

    node_hierarchy = key.split('.')
    # first, search on current node
    # second, search on root if search failed on current node

    node, unpacked_key = search_node(node_hierarchy, node)
    if node and unpacked_key:
        return node, unpacked_key

    node = root
    node, unpacked_key = search_node(node_hierarchy, node)
    if node and unpacked_key:
        return node, unpacked_key
    raise ValueError(f'config doesn\'t contain key: {key}')


def search_node(node_hierarchy, node):
    for i, key in enumerate(node_hierarchy[::-1]):
        if key in node:
            if i == len(node_hierarchy) - 1:
                return node, node_hierarchy[-1]
            else:
                node = node[key]
    return None, None



import numpy as np
from einops import rearrange, repeat

from jbag.transforms.transforms import Transform


class ToType(Transform):
    def __init__(self, keys, dtype):
        super().__init__(keys)
        self.dtype = dtype

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key].astype(self.dtype)
            data[key] = value
        return data


class Rearrange(Transform):
    def __init__(self, keys, pattern):
        """
        Change the arrangement of given elements.

        Args:
            keys (str or sequence):
            pattern (str): Arranging pattern. For example "i j k -> j k i".
        """

        super().__init__(keys)
        self.pattern = pattern

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key]
            value = rearrange(value, self.pattern)
            data[key] = value
        return data


class Repeat(Transform):
    def __init__(self, keys, pattern, **kwargs):
        super().__init__(keys)
        self.pattern = pattern
        self.kwargs = kwargs

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key]
            value = repeat(value, self.pattern, **self.kwargs)
            data[key] = value
        return data


class AddChannel(Transform):
    def __init__(self, keys, dim):
        """
        Add additional dimension in specific position.

        Args:
            keys (str or sequence):
            dim (int):
        """

        super().__init__(keys)
        self.dim = dim

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key]
            value = np.expand_dims(value, axis=self.dim)
            data[key] = value
        return data


class ZscoreNormalization(Transform):
    def __init__(self, keys, mean_value=None, std_value=None, mean_key=None, std_key=None):
        """
        Perform z-score normalization. `mean_key` and `std_key` indicate keys of mean and std value in data. You can
        also set common mean and std values for all data. Mean and std values provided by each sample will be used
        firstly if they exist.

        Args:
            keys (str or sequence):
            mean_value (float or None, optional, default=None):
            std_value (float or None, optional, default=None):
            mean_key (str or None, optional, default=None):
            std_key (str or None, optional, default=None):
        """

        super().__init__(keys)
        self.mean_value = mean_value
        self.std_value = std_value
        self.mean_key = mean_key
        self.std_key = std_key

    def _call_fun(self, data):
        mean = data[self.mean_key] if self.mean_key in data else self.mean_value
        std = data[self.std_key] if self.std_key in data else self.std_value
        assert mean and std

        for key in self.keys:
            value = data[key]
            value = (value - mean) / std
            data[key] = value
        return data


class MinMaxNormalization(Transform):
    def __init__(self, keys, lower_bound_percentile=1, upper_bound_percentile=99):
        """
        Perform min-max normalization.

        Args:
            keys (str or sequence):
            lower_bound_percentile (int, optional, default=1):
            upper_bound_percentile (int, optional, default=99):
        """

        super().__init__(keys)
        self.lower_bound_percentile = lower_bound_percentile
        self.upper_bound_percentile = upper_bound_percentile

    def _call_fun(self, data):
        for key in self.keys:
            image = data[key]
            min_value, max_value = np.percentile(image, (self.lower_bound_percentile, self.upper_bound_percentile))
            image = (image - min_value) / (max_value - min_value)
            data[key] = image
        return data



import torch
import torch.nn.functional as F
from torch import nn


class DoubleConv(nn.Sequential):
    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels

        conv1 = nn.Conv2d(in_channels=in_channels, out_channels=mid_channels, kernel_size=3, padding=1)
        bn1 = nn.BatchNorm2d(num_features=mid_channels)
        relu1 = nn.ReLU(inplace=True)
        conv2 = nn.Conv2d(in_channels=mid_channels, out_channels=out_channels, kernel_size=3, padding=1)
        bn2 = nn.BatchNorm2d(num_features=out_channels)
        relu2 = nn.ReLU(inplace=True)
        super().__init__(conv1, bn1, relu1, conv2, bn2, relu2)


class UNetEncoder(nn.Module):
    def __init__(self, in_channels=1, width_factor=64, blocks=5):
        super().__init__()
        channels = [width_factor << i for i in range(blocks)]

        block_0 = DoubleConv(in_channels=in_channels, out_channels=channels[0])
        self.blocks = nn.ModuleList([
            nn.Sequential(nn.MaxPool2d(2), DoubleConv(in_channels=channels[i - 1], out_channels=channels[i]))
            for i in range(1, blocks)])

        self.blocks.insert(0, block_0)
        self.out_channels = channels

    def forward(self, x):
        features = []
        for block in self.blocks:
            x = block(x)
            features.append(x)
        return features


class UNetDecoder(nn.Module):
    def __init__(self, encoder_channels):
        super().__init__()

        in_channels = encoder_channels[::-1]

        self.blocks = nn.ModuleList([
            Up(in_channels=in_channels[i], out_channels=in_channels[i + 1],
               skip_conn_channels=in_channels[i + 1])
            for i in range(0, len(in_channels) - 1)])

        self.out_channels = in_channels[-1]

    def forward(self, x):
        skip_connections = x[-2::-1]
        x = x[-1]
        for i, skip_connection in enumerate(skip_connections):
            x = self.blocks[i](x, skip_connection)
        return x


class Up(nn.Module):
    def __init__(self, in_channels, out_channels, skip_conn_channels):
        super().__init__()
        self.up = nn.ConvTranspose2d(in_channels, in_channels // 2, kernel_size=2, stride=2)
        self.conv = DoubleConv(in_channels // 2 + skip_conn_channels, out_channels)

    def forward(self, x, skip_features):
        x = self.up(x)
        diff_y = skip_features.size()[2] - x.size()[2]
        diff_x = skip_features.size()[3] - x.size()[3]
        if diff_y != 0 or diff_x != 0:
            x = F.pad(x, [diff_x // 2, diff_x - diff_x // 2,
                          diff_y // 2, diff_y - diff_y // 2])

        x = torch.cat([skip_features, x], dim=1)
        return self.conv(x)


class UNet(nn.Module):
    def __init__(self, in_channels=1, out_channels=1, width_factor=64, blocks=5, normalize=True):
        """
        UNet.

        Args:
            in_channels (int, optional, default=1):
            out_channels (int, optional, default=1):
            width_factor (int, optional, default=64):
            blocks (int, optional, default=5):
            normalize (bool, optional, default=True): If `True`, normalize the output using `torch.sigmoid` for one
                dimension output, or `torch.softmax` for multiple classes output.
        """

        super().__init__()
        self.out_channels = out_channels
        encoder = UNetEncoder(in_channels=in_channels, width_factor=width_factor, blocks=blocks)
        decoder = UNetDecoder(encoder_channels=encoder.out_channels)
        decoder_head = nn.Conv2d(in_channels=decoder.out_channels, out_channels=out_channels, kernel_size=1)
        self.unet = nn.Sequential(encoder,
                                  decoder,
                                  decoder_head)

    def forward(self, x):
        x = self.unet(x)
        return x
